/***************************************************
* By XC8 compiler
*  File Name: remote_ctrl.c
*  Description: learnign remote control
*  RA0     not use(ICSPDAT)
*  RA1     not use(ICSPCLK)
*  RA2 in  VOUT RC Receiver
*  RA3     not use(MCLR)
*  RA4     not use
*  RA5     not use
*  RC0     not use
*  RC1 out Relay Ctrl 
*  RC2     not use
*  RC3 in  SET SW
*  RC4     not use
*  RC5 OUT LED
*  COPYRIGHT 2016 MYCOMKITS.COM, owned by CNET LIMITED
*  vO̒쌠́Aҁu}CRLbghbgR^c@LЃNlbgvɋA܂B
*  쌠Ă܂񂪁AvOgwK̒ŃvORɕύXĂgB
***************************************************/


#include <xc.h>

#define _XTAL_FREQ  8000000  // Clock Frequency 8MHz

// #pragma config statements should precede project file includes.
// Use project enums instead of #define for ON and OFF.

// CONFIG1
#pragma config FOSC = INTOSC    // Oscillator Selection (INTOSC oscillator: I/O function on CLKIN pin)
#pragma config WDTE = OFF       // Watchdog Timer Enable (WDT disabled)
#pragma config PWRTE = ON       // Power-up Timer Enable (PWRT enabled)
#pragma config MCLRE = OFF      // MCLR Pin Function Select (MCLR/VPP pin function is digital input)
#pragma config CP = OFF         // Flash Program Memory Code Protection (Program memory code protection is disabled)
#pragma config CPD = OFF        // Data Memory Code Protection (Data memory code protection is disabled)
#pragma config BOREN = ON       // Brown-out Reset Enable (Brown-out Reset enabled)
#pragma config CLKOUTEN = OFF   // Clock Out Enable (CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin)
#pragma config IESO = OFF       // Internal/External Switchover (Internal/External Switchover mode is disabled)
#pragma config FCMEN = OFF      // Fail-Safe Clock Monitor Enable (Fail-Safe Clock Monitor is disabled)

// CONFIG2
#pragma config WRT = OFF        // Flash Memory Self-Write Protection (Write protection off)
#pragma config PLLEN = OFF      // PLL Enable (4x PLL disabled)
#pragma config STVREN = OFF     // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will not cause a Reset)
#pragma config BORV = LO        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), low trip point selected.)
#pragma config LVP = OFF        // Low-Voltage Programming Enable (High-voltage on MCLR/VPP must be used for programming)


// Funtioction
extern int GetPattern(void);
extern int MatchPattern(void);
extern void SavePattern(void);
extern void RecallPattern(void);
extern void LedOk(void);
extern void LedErr(void);


#define LEN_MAX 250

// Variables
short interval_1;
short cnt_1;
short cnt_time;
int rc_mode;
int rc_err;
int mem_err;
int relay_ctrl;

unsigned char bmap[LEN_MAX+1];
unsigned char rc_len;
unsigned char rc_pat[LEN_MAX];
unsigned char mem_len;
unsigned char mem_pat[LEN_MAX];

/*
 * Main Routine
 */
int main(int argc, char** argv) {

    // Clock Setup
    OSCCON = 0x70;  // 8MHz
  
    //I/O Port Setup
    OPTION_REGbits.nWPUEN = 0;  // Weak pullup enabled
    TRISA = 0x3F;   // All Input
    ANSELA = 0x00;	// All Digital
    WPUA = 0x3B;    // RA5-RA3,RA1-RA0 pullup
    TRISC = 0x1D;	// RC5,RC1 Output
    ANSELC = 0x00;  // All Digital
    WPUC = 0x1D;    // RC4-RC2,RC0 pullup
  
    // Timer1 Setup
    T1CON = 0x51;       // Fosc, 1:2, Enables Timer1
    T1GCON = 0x00;          // Gate disabled
    TMR1L = 0x00;       // about 1/61sec 0.016384sec
    TMR1H = 0x00;       // 
    PIR1bits.TMR1IF = 0;   // Timer1 Flag Clear 
    interval_1 = 0;    // about 1/61sec 0.016384sec
    cnt_1 = 0;
    cnt_time = 300;      // Interval = 1/61sec*(cnt_time + 1)
    
    relay_ctrl = 0;
    PORTCbits.RC1 = relay_ctrl;  // Relay OFF
    
    PORTCbits.RC5 = 1;
    __delay_ms(1000);
    PORTCbits.RC5 = 0;

    rc_mode = 0;
    
    RecallPattern();
    
    //
    // Main routine
    //
    while(1){
        if(rc_mode == 0){  // idle
            if(PORTCbits.RC3 == 0){ // SET SW Pushed
                __delay_ms(100);
                PORTCbits.RC5 = 1;
                TMR1L = 0x00;       // about 1/4sec 0.2621sec
                TMR1H = 0x00;       // 
                PIR1bits.TMR1IF = 0;   // Timer1 Flag Clear
                cnt_1 = 0;
                rc_mode = 1;   // Wait RC pattern for Memory
            }
            if(PORTAbits.RA2 == 0){  // Vout Low
                rc_mode = 2;   //  Get RC pattern for RELAY
            }
        }else if(rc_mode == 1){    // Wait RC pattern for Memory
            if(PORTAbits.RA2 == 0){  // Vout Low
                rc_err = GetPattern();   //  Get RC pattern
                if(rc_err == 0){
                    SavePattern();
                    LedOk();
                }else{
                    LedErr();
                }
                rc_mode = 0;    // Idle
            }else if(PIR1bits.TMR1IF == 1){   // Timer1 Inerval
                if(cnt_1 < cnt_time){
                    TMR1L = interval_1;
                    TMR1H = interval_1 >> 8;
                    PIR1bits.TMR1IF = 0;    // Timer1 Flag Clear 
                    cnt_1++;
                }else{
                    cnt_1 = 0;
                    rc_mode = 4;    // Timeout waiting RC pattern
                }
            }
        }else if(rc_mode == 2){     //  Get RC pattern for RELAY
                rc_err = GetPattern();   //  Get RC pattern
                if(rc_err == 0){
                    mem_err = MatchPattern();
                    if(mem_err == 0){
                        relay_ctrl = (~relay_ctrl) & 0x01;
                        PORTCbits.RC1 = relay_ctrl; // Relay Toggle   
                    }
                    __delay_ms(200);
                }
                rc_mode = 0;    // idle
        }else if(rc_mode == 4){     // Timeout waiting RC pattern
            LedErr();
            rc_mode = 0;    // idle
        }
    }
}

//
// Pattern Function
//
int GetPattern(){
    int i;
    
    TMR1L = 0;  // clear timer 1
    TMR1H = 0;  // clear timer 1

    while(1){   // detect "1" length for leader
        if(PORTAbits.RA2 == 1){
            rc_pat[0] = TMR1H;
            TMR1L = 0;  // clear timer 1
            TMR1H = 0;  // clear timer 1
            break;
        }
        if(TMR1H > 200){    // about 12ms
            return 1;   // timeout error
        }
    }
    
    while(1){   // detect "0" length for leader
        if(PORTAbits.RA2 == 0){
            rc_pat[1] = TMR1H;  // 1 count = 64us
            TMR1L = 0;  // clear timer 1
            TMR1H = 0;  // clear timer 1
            break;
        }
        if(TMR1H > 100){    // about 6ms
            return 1;   // timeout error
        }
    }
    
    for(i=1;i<(LEN_MAX/2);i++){
        while(1){   // detect "1" length
            if(PORTAbits.RA2 == 1){
                rc_pat[2*i] = TMR1H;
                TMR1L = 0;  // clear timer 1
                TMR1H = 0;  // clear timer 1
                break;
            }
            if(TMR1H > 100){    // about 6ms
                return 1;   // timeout error
            }
        }

        while(1){   // detect "0" length
            if(PORTAbits.RA2 == 0){
                rc_pat[2*i+1] = TMR1H;
                TMR1L = 0;  // clear timer 1
                TMR1H = 0;  // clear timer 1
                break;
            }
            if(TMR1H > 100){    // about 6ms
                rc_len = 2*i+1;
                return 0;   // pattern end
            }
        }
    }
    
    return 1;   // too long pattern error
}

int MatchPattern(){
    int i;
    
    
    if(mem_len != rc_len) return 1; // check pattern length

    // check Leader pattern
    if((mem_pat[0] > (rc_pat[0] + 16)) || ((mem_pat[i] + 16) < rc_pat[i])) return 1;  // Leader HIGH length mergin 1000us
    if((mem_pat[0] > (rc_pat[0] + 8)) || ((mem_pat[i] + 8) < rc_pat[i])) return 1;  // Leader LOW length  mergin 500us

    // chech data pattern
    for(i=2;i<rc_len;i++){
        if((mem_pat[i] > (rc_pat[i] + 4)) || ((mem_pat[i] + 4) < rc_pat[i])) return 1;  // mergin 250us
    }
    
    return 0;   // pattern matched
}

void SavePattern(){
    int i;
    
    // copy receive pattern
    mem_len = rc_len;
    for(i=0;i<rc_len;i++){
        mem_pat[i] = rc_pat[i];
    }
        
    // Save to EEPROM
    eeprom_write(0,mem_len);
    for(i=1;i<(LEN_MAX+1);i++) eeprom_write(i,mem_pat[i-1]);
}

void RecallPattern(){
    int i;
    
    // Read from EEPROM
    mem_len = eeprom_read(0);
    for(i=1;i<(LEN_MAX+1);i++) mem_pat[i-1] = eeprom_read(i);
}

//
// LED brink
//
void LedOk(void){
    PORTCbits.RC5 = 0;
    __delay_ms(500);
    PORTCbits.RC5 = 1;
    __delay_ms(100);
    PORTCbits.RC5 = 0;
}

void LedErr(void){
    PORTCbits.RC5 = 0;
    __delay_ms(500);
    PORTCbits.RC5 = 1;
    __delay_ms(300);
    PORTCbits.RC5 = 0;
    __delay_ms(300);
    PORTCbits.RC5 = 1;
    __delay_ms(300);
    PORTCbits.RC5 = 0;
    __delay_ms(300);
    PORTCbits.RC5 = 1;
    __delay_ms(300);
    PORTCbits.RC5 = 0;
}
